# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.parsers
~~~~~~~~~~~~~~~~~~~

This module contains core API parsers.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError

from .compat import unquote
from .credentials import Value
from .definitions import (Bucket, CopyObjectResult, IncompleteUpload,
                          MultipartUploadResult, Object, UploadPart)
# minio specific.
from .error import InvalidXMLError, MultiDeleteError
from .helpers import _iso8601_to_utc_datetime
from .xml_marshal import NOTIFICATIONS_ARN_FIELDNAME_MAP

# dependencies.


_XML_NS = {
    's3': 'http://s3.amazonaws.com/doc/2006-03-01/',
    'sts': 'https://sts.amazonaws.com/doc/2011-06-15/'
}


class S3Element:
    """S3 aware XML parsing class. Wraps a root element name and
    ElementTree.Element instance. Provides S3 namespace aware parsing
    functions.

    """

    def __init__(self, root_name, element):
        self.root_name = root_name
        self.element = element

    @classmethod
    def fromstring(cls, root_name, data):
        """Initialize S3Element from name and XML string data.

        :param name: Name for XML data. Used in XML errors.
        :param data: string data to be parsed.
        :return: Returns an S3Element.
        """
        try:
            return cls(root_name, ElementTree.fromstring(data.strip()))
        except (ParseError, AttributeError, ValueError, TypeError) as error:
            raise InvalidXMLError(
                '"{}" XML is not parsable. Message: {}'.format(
                    root_name, error
                )
            )

    def findall(self, name):
        """Similar to ElementTree.Element.findall()

        """
        return [
            S3Element(self.root_name, elem)
            for elem in self.element.findall('s3:{}'.format(name), _XML_NS)
        ]

    def find(self, name):
        """Similar to ElementTree.Element.find()

        """
        elt = self.element.find('s3:{}'.format(name), _XML_NS)
        return S3Element(self.root_name, elt) if elt else None

    def get_child_text(self, name, strict=True):
        """Extract text of a child element. If strict, and child element is
        not present, raises InvalidXMLError and otherwise returns
        None.

        """
        if strict:
            try:
                return self.element.find('s3:{}'.format(name), _XML_NS).text
            except (ParseError, AttributeError, ValueError, TypeError) as error:
                raise InvalidXMLError(
                    ('Invalid XML provided for "{}" - erroring tag <{}>. '
                     'Message: {}').format(self.root_name, name, error)
                )
        else:
            return self.element.findtext('s3:{}'.format(name), None, _XML_NS)

    def get_urldecoded_elem_text(self, name, strict=True):
        """Like self.get_child_text(), but also performs urldecode() on the
        result.

        """
        text = self.get_child_text(name, strict)
        # strictness is already enforced above.
        return unquote(text) if text is not None else None

    def get_etag_elem(self, strict=True):
        """Fetches an 'ETag' child element suitably processed.

        """
        return self.get_child_text('ETag', strict).replace('"', '')

    def get_int_elem(self, name):
        """Fetches an integer type XML child element by name.

        """
        return int(self.get_child_text(name))

    def get_localized_time_elem(self, name):
        """Parse a time XML child element.

        """
        return _iso8601_to_utc_datetime(self.get_child_text(name))

    def text(self):
        """Fetch the current node's text

        """
        return self.element.text

    def is_dir(self):
        """Returns True if the object is a dir
        ie, if an object name has `/` suffixed.

        """
        text = self.get_child_text('Key')
        return text.endswith("/")


def parse_multipart_upload_result(data):
    """
    Parser for complete multipart upload response.

    :param data: Response data for complete multipart upload.
    :return: :class:`MultipartUploadResult <MultipartUploadResult>`.
    """
    root = S3Element.fromstring('CompleteMultipartUploadResult', data)

    return MultipartUploadResult(
        root.get_child_text('Bucket'),
        root.get_child_text('Key'),
        root.get_child_text('Location'),
        root.get_etag_elem()
    )


def parse_copy_object(bucket_name, object_name, data):
    """
    Parser for copy object response.

    :param data: Response data for copy object.
    :return: :class:`CopyObjectResult <CopyObjectResult>`
    """
    root = S3Element.fromstring('CopyObjectResult', data)

    return CopyObjectResult(
        bucket_name, object_name,
        root.get_etag_elem(),
        root.get_localized_time_elem('LastModified')
    )


def parse_list_buckets(data):
    """
    Parser for list buckets response.

    :param data: Response data for list buckets.
    :return: List of :class:`Bucket <Bucket>`.
    """
    root = S3Element.fromstring('ListBucketsResult', data)

    return [
        Bucket(bucket.get_child_text('Name'),
               bucket.get_localized_time_elem('CreationDate'))
        for buckets in root.findall('Buckets')
        for bucket in buckets.findall('Bucket')
    ]


def _parse_objects_from_xml_elts(bucket_name, contents, common_prefixes,
                                 delete_markers=()):
    """Internal function that extracts objects and common prefixes from
    list_objects responses.
    """
    objects = [
        Object(
            bucket_name,
            content.get_child_text("Key"),
            last_modified=content.get_localized_time_elem("LastModified"),
            etag=content.get_etag_elem(strict=False),
            size=content.get_int_elem("Size"),
            is_dir=content.is_dir(),
            version_id=content.get_child_text("VersionId", strict=False),
            is_latest=content.get_child_text("IsLatest", strict=False),
            storage_class=content.get_child_text("StorageClass", strict=False),
            owner_id=(
                content.find("Owner").get_child_text("ID", strict=False)
                if content.find("Owner") else None
            ),
            owner_name=(
                content.find("Owner").get_child_text(
                    "DisplayName", strict=False,
                ) if content.find("Owner") else None
            ),
        ) for content in contents
    ]

    object_dirs = [
        Object(bucket_name, dir_elt.text(), is_dir=True)
        for dirs_elt in common_prefixes
        for dir_elt in dirs_elt.findall('Prefix')
    ]

    markers = [
        Object(
            bucket_name,
            content.get_child_text("Key"),
            last_modified=content.get_localized_time_elem("LastModified"),
            is_dir=content.is_dir(),
            version_id=content.get_child_text("VersionId", strict=False),
            is_latest=content.get_child_text("IsLatest", strict=False),
            owner_id=(
                content.find("Owner").get_child_text("ID", strict=False)
                if content.find("Owner") else None
            ),
            owner_name=(
                content.find("Owner").get_child_text(
                    "DisplayName", strict=False,
                ) if content.find("Owner") else None
            ),
            delete_marker=True,
        ) for content in delete_markers
    ]

    return objects, object_dirs, markers


def parse_list_objects(data, bucket_name):
    """
    Parser for list objects response.

    :param data: Response data for list objects.
    :param bucket_name: Response for the bucket.
    :return: Replies back three distinctive components.
       - List of :class:`Object <Object>`
       - True if list is truncated, False otherwise.
       - Object name marker for the next request.
    """
    root = S3Element.fromstring('ListObjectResult', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    # NextMarker element need not be present.
    marker = root.get_urldecoded_elem_text('NextMarker', strict=False)
    objects, object_dirs, _ = _parse_objects_from_xml_elts(
        bucket_name,
        root.findall('Contents'),
        root.findall('CommonPrefixes')
    )

    if is_truncated and marker is None:
        marker = objects[-1].object_name

    return objects + object_dirs, is_truncated, marker


def parse_list_objects_v2(data, bucket_name):
    """
    Parser for list objects version 2 response.

    :param data: Response data for list objects.
    :param bucket_name: Response for the bucket.
    :return: Returns three distinct components:
       - List of :class:`Object <Object>`
       - True if list is truncated, False otherwise.
       - Continuation Token for the next request.
    """
    root = S3Element.fromstring('ListObjectV2Result', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    # NextContinuationToken may not be present.
    continuation_token = root.get_child_text('NextContinuationToken',
                                             strict=False)
    objects, object_dirs, _ = _parse_objects_from_xml_elts(
        bucket_name,
        root.findall('Contents'),
        root.findall('CommonPrefixes')
    )

    return objects + object_dirs, is_truncated, continuation_token


def parse_list_object_versions(data, bucket_name):
    """
    Parser for list object versions response.

    :param data: Response data for list objects.
    :param bucket_name: Response for the bucket.
    :return: Returns three distinct components:
       - List of :class:`Object <Object>`
       - True if list is truncated, False otherwise.
       - Continuation Token for the next request.
    """
    root = S3Element.fromstring("ListVersionsResult", data)

    is_truncated = root.get_child_text("IsTruncated").lower() == "true"

    key_marker = root.get_urldecoded_elem_text("NextKeyMarker", strict=False)
    version_id_marker = root.get_urldecoded_elem_text(
        "NextVersionIdMarker",
        strict=False,
    )

    objects, object_dirs, delete_markers = _parse_objects_from_xml_elts(
        bucket_name,
        root.findall("Version"),
        root.findall("CommonPrefixes"),
        root.findall("DeleteMarker"),
    )

    return (
        objects + object_dirs + delete_markers,
        is_truncated,
        key_marker,
        version_id_marker,
    )


def parse_list_multipart_uploads(data, bucket_name):
    """
    Parser for list multipart uploads response.

    :param data: Response data for list multipart uploads.
    :param bucket_name: Response for the bucket.
    :return: Replies back four distinctive components.
       - List of :class:`IncompleteUpload <IncompleteUpload>`
       - True if list is truncated, False otherwise.
       - Object name marker for the next request.
       - Upload id marker for the next request.
    """
    root = S3Element.fromstring('ListMultipartUploadsResult', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    key_marker = root.get_urldecoded_elem_text('NextKeyMarker', strict=False)
    upload_id_marker = root.get_child_text('NextUploadIdMarker', strict=False)
    uploads = [
        IncompleteUpload(bucket_name,
                         upload.get_urldecoded_elem_text('Key'),
                         upload.get_child_text('UploadId'),
                         upload.get_localized_time_elem('Initiated'))
        for upload in root.findall('Upload')
    ]

    return uploads, is_truncated, key_marker, upload_id_marker


def parse_list_parts(data, bucket_name, object_name, upload_id):
    """
    Parser for list parts response.

    :param data: Response data for list parts.
    :param bucket_name: Response for the bucket.
    :param object_name: Response for the object.
    :param upload_id: Upload id of object name for
       the active multipart session.
    :return: Replies back three distinctive components.
       - List of :class:`UploadPart <UploadPart>`.
       - True if list is truncated, False otherwise.
       - Next part marker for the next request if the
         list was truncated.
    """
    root = S3Element.fromstring('ListPartsResult', data)

    is_truncated = root.get_child_text('IsTruncated').lower() == 'true'
    part_marker = root.get_child_text('NextPartNumberMarker', strict=False)
    parts = [
        UploadPart(bucket_name, object_name, upload_id,
                   part.get_int_elem('PartNumber'),
                   part.get_etag_elem(),
                   part.get_localized_time_elem('LastModified'),
                   part.get_int_elem('Size'))
        for part in root.findall('Part')
    ]

    return parts, is_truncated, part_marker


def parse_new_multipart_upload(data):
    """
    Parser for new multipart upload response.

    :param data: Response data for new multipart upload.
    :return: Returns a upload id.
    """
    root = S3Element.fromstring('InitiateMultipartUploadResult', data)
    return root.get_child_text('UploadId')


def parse_location_constraint(data):
    """
    Parser for location constraint response.

    :param data: Response data for get bucket location.
    :return: Returns location of your bucket.
    """
    root = S3Element.fromstring('BucketLocationConstraintResult', data)
    return root.text()


def parse_get_bucket_notification(data):
    """
    Parser for a get_bucket_notification response from S3.

    :param data: Body of response from get_bucket_notification.
    :return: Returns bucket notification configuration
    """
    root = S3Element.fromstring('GetBucketNotificationResult', data)

    notifications = _add_notifying_service_config(
        root, {},
        'TopicConfigurations', 'TopicConfiguration'
    )
    notifications = _add_notifying_service_config(
        root, notifications,
        'QueueConfigurations', 'QueueConfiguration'
    )
    notifications = _add_notifying_service_config(
        root, notifications,
        'CloudFunctionConfigurations', 'CloudFunctionConfiguration'
    )

    return notifications


def _add_notifying_service_config(data, notifications, service_key,
                                  service_xml_tag):
    """Add service configuration in notification."""

    arn_elt_name = NOTIFICATIONS_ARN_FIELDNAME_MAP[service_xml_tag]
    config = []
    for service in data.findall(service_xml_tag):
        config_item = {}
        config_item['Id'] = service.get_child_text('Id')
        config_item['Arn'] = service.get_child_text(arn_elt_name)
        config_item['Events'] = [
            event.text() for event in service.findall('Event')
        ]
        filter_terms = [
            {
                'Key': {
                    'FilterRules': [
                        {
                            'Name': xml_filter_rule.get_child_text('Name'),
                            'Value': xml_filter_rule.get_child_text('Value'),
                        }
                        for xml_filter_rule in xml_filter_rules.findall(
                            './S3Key/FilterRule')
                    ]
                }
            }
            for xml_filter_rules in service.findall('Filter')
        ]
        if len(filter_terms) > 0:
            config_item['Filter'] = filter_terms
        config.append(config_item)

    if len(config) > 0:
        notifications[service_key] = config

    return notifications


def parse_multi_delete_response(data):
    """Parser for Multi-Object Delete API response.

    :param data: XML response body content from service.

    :return: Returns list of error objects for each delete object that
    had an error.

    """
    root = S3Element.fromstring('MultiObjectDeleteResult', data)
    return [
        MultiDeleteError(errtag.get_child_text('Key'),
                         errtag.get_child_text('Code'),
                         errtag.get_child_text('Message'))
        for errtag in root.findall('Error')
    ]


def parse_assume_role(data):
    """
    Parser for assume role response.

    :param data: XML response data for STS assume role as a string.
    :return: A 2-tuple containing:
        - a :class:`minio.credentials.Value` instance with the
                    temporary credentials.
        - A :class:`DateTime` instance of when the credentials expire.
    """
    root = ElementTree.fromstring(data)
    credentials = root.find("sts:AssumeRoleResult", _XML_NS).find(
        "sts:Credentials", _XML_NS)

    access_key = credentials.find("sts:AccessKeyId", _XML_NS).text
    secret_key = credentials.find("sts:SecretAccessKey", _XML_NS).text
    session_token = credentials.find("sts:SessionToken", _XML_NS).text

    expiry_str = credentials.find("sts:Expiration", _XML_NS).text
    expiry = _iso8601_to_utc_datetime(expiry_str)

    return Value(access_key, secret_key, session_token), expiry



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016, 2017, 2018, 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.xml_marshal
~~~~~~~~~~~~~~~

This module contains the simple wrappers for XML marshaller's.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import

import io
from collections import defaultdict
from xml.etree import ElementTree as ET

from .compat import basestring

_S3_NAMESPACE = 'http://s3.amazonaws.com/doc/2006-03-01/'


def Element(tag, with_namespace=False):  # pylint: disable=invalid-name
    """Create ElementTree.Element with tag and namespace."""
    if with_namespace:
        return ET.Element(tag, {'xmlns': _S3_NAMESPACE})
    return ET.Element(tag)


def SubElement(parent, tag, text=None):  # pylint: disable=invalid-name
    """Create ElementTree.SubElement on parent with tag and text."""
    element = ET.SubElement(parent, tag)
    if text is not None:
        element.text = text
    return element


def _get_xml_data(element):
    """Get XML data of ElementTree.Element."""
    data = io.BytesIO()
    ET.ElementTree(element).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()


def _etree_to_dict(elem):
    """Converts ElementTree object to dict."""
    ns = '{' + _S3_NAMESPACE + '}'  # pylint: disable=invalid-name
    elem.tag = elem.tag.replace(ns, '')

    d = {elem.tag: {} if elem.attrib else None}  # pylint: disable=invalid-name
    children = list(elem)
    if children:
        dd = defaultdict(list)  # pylint: disable=invalid-name
        is_rule = children[0].tag.replace(ns, "") == "Rule"
        # pylint: disable=invalid-name
        for dc in map(_etree_to_dict, children):
            for k, v in dc.items():  # pylint: disable=invalid-name
                dd[k].append([v] if is_rule else v)
        # pylint: disable=invalid-name
        d = {elem.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if elem.attrib:
        d[elem.tag].update(('@' + k, v) for k, v in elem.attrib.items())
    if elem.text:
        text = elem.text.strip()
        if children or elem.attrib:
            if text:
                d[elem.tag]['#text'] = text
        else:
            d[elem.tag] = text
    return d


def xml_to_dict(in_xml):
    """Convert XML to dict."""
    elem = ET.XML(in_xml)
    return _etree_to_dict(elem)


def xml_marshal_bucket_encryption(rules):
    """Encode bucket encryption to XML."""

    root = Element('ServerSideEncryptionConfiguration')

    if rules:
        # As server supports only one rule, the first rule is taken due to
        # no validation is done at server side.
        apply_element = SubElement(SubElement(root, 'Rule'),
                                   'ApplyServerSideEncryptionByDefault')
        SubElement(apply_element, 'SSEAlgorithm',
                   rules[0]['ApplyServerSideEncryptionByDefault'].get(
                       'SSEAlgorithm', 'AES256'))
        kms_text = rules[0]['ApplyServerSideEncryptionByDefault'].get(
            'KMSMasterKeyID')
        if kms_text:
            SubElement(apply_element, 'KMSMasterKeyID', kms_text)

    return _get_xml_data(root)


def xml_marshal_bucket_constraint(region):
    """
    Marshal's bucket constraint based on *region*.

    :param region: Region name of a given bucket.
    :return: Marshalled XML data.
    """
    root = Element('CreateBucketConfiguration', with_namespace=True)
    SubElement(root, 'LocationConstraint', region)
    return _get_xml_data(root)


def xml_marshal_select(req):
    """Encode select request to XML."""

    def bool_to_str(value):
        return "true" if value else "false"

    root = Element("SelectObjectContentRequest")
    SubElement(root, "Expression", req.expression)
    SubElement(root, "ExpressionType", "SQL")

    input_serialization = SubElement(root, "InputSerialization")
    SubElement(
        input_serialization,
        "CompressionType",
        req.input_serialization.compression_type,
    )

    if req.input_serialization.csv:
        csv = SubElement(input_serialization, "CSV")
        SubElement(
            csv,
            "FileHeaderInfo",
            req.input_serialization.csv.file_header_info,
        )
        SubElement(
            csv,
            "RecordDelimiter",
            req.input_serialization.csv.record_delimiter,
        )
        SubElement(
            csv, "FieldDelimiter", req.input_serialization.csv.field_delimiter,
        )
        SubElement(
            csv, "QuoteCharacter", req.input_serialization.csv.quote_character,
        )
        SubElement(
            csv,
            "QuoteEscapeCharacter",
            req.input_serialization.csv.quote_escape_character,
        )
        SubElement(csv, "Comments", req.input_serialization.csv.comments)
        SubElement(
            csv,
            "AllowQuotedRecordDelimiter",
            bool_to_str(
                req.input_serialization.csv.allow_quoted_record_delimiter,
            ),
        )

    if req.input_serialization.json:
        SubElement(
            SubElement(input_serialization, "JSON"),
            "Type",
            req.input_serialization.json.json_type,
        )

    if req.input_serialization.parquet:
        SubElement(input_serialization, "Parquet")

    output_serialization = SubElement(root, "OutputSerialization")
    if req.output_serialization.csv:
        csv = SubElement(output_serialization, "CSV")
        SubElement(
            csv,
            "QuoteFields",
            req.output_serialization.csv.quote_fields,
        )
        SubElement(
            csv,
            "RecordDelimiter",
            req.output_serialization.csv.record_delimiter,
        )
        SubElement(
            csv,
            "FieldDelimiter",
            req.output_serialization.csv.field_delimiter,
        )
        SubElement(
            csv,
            "QuoteCharacter",
            req.output_serialization.csv.quote_character,
        )
        SubElement(
            csv,
            "QuoteEscapeCharacter",
            req.output_serialization.csv.quote_escape_character,
        )

    if req.output_serialization.json:
        SubElement(
            SubElement(output_serialization, "JSON"),
            "RecordDelimiter",
            req.output_serialization.json.record_delimiter,
        )

    SubElement(
        SubElement(root, "RequestProgress"),
        "Enabled",
        bool_to_str(req.request_progress.enabled),
    )

    return _get_xml_data(root)


def marshal_complete_multipart(uploaded_parts):
    """
    Marshal's complete multipart upload request based on *uploaded_parts*.

    :param uploaded_parts: List of all uploaded parts, ordered by part number.
    :return: Marshalled XML data.
    """
    root = Element('CompleteMultipartUpload', with_namespace=True)
    for uploaded_part in uploaded_parts:
        part = SubElement(root, 'Part')
        SubElement(part, 'PartNumber', str(uploaded_part.part_number))
        SubElement(part, 'ETag', '"' + uploaded_part.etag + '"')

    return _get_xml_data(root)


def marshal_bucket_notifications(notifications):
    """
    Marshals the notifications structure for sending to S3 compatible storage

    :param notifications: Dictionary with following structure:

    {
        'TopicConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|
                    's3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|
                    's3:ObjectCreated:Copy'|
                    's3:ObjectCreated:CompleteMultipartUpload'|
                    's3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|
                    's3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ],
        'QueueConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|
                    's3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|
                    's3:ObjectCreated:Copy'|
                    's3:ObjectCreated:CompleteMultipartUpload'|
                    's3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|
                    's3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ],
        'CloudFunctionConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|
                    's3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|
                    's3:ObjectCreated:Copy'|
                    's3:ObjectCreated:CompleteMultipartUpload'|
                    's3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|
                    's3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ]
    }

    :return: Marshalled XML data
    """
    root = Element('NotificationConfiguration', with_namespace=True)
    _add_notification_config_to_xml(
        root,
        'TopicConfiguration',
        notifications.get('TopicConfigurations', [])
    )
    _add_notification_config_to_xml(
        root,
        'QueueConfiguration',
        notifications.get('QueueConfigurations', [])
    )
    _add_notification_config_to_xml(
        root,
        'CloudFunctionConfiguration',
        notifications.get('CloudFunctionConfigurations', [])
    )

    return _get_xml_data(root)


NOTIFICATIONS_ARN_FIELDNAME_MAP = {
    'TopicConfiguration': 'Topic',
    'QueueConfiguration': 'Queue',
    'CloudFunctionConfiguration': 'CloudFunction',
}


def _add_notification_config_to_xml(node, element_name, configs):
    """
    Internal function that builds the XML sub-structure for a given
    kind of notification configuration.

    """
    for config in configs:
        config_node = SubElement(node, element_name)

        if 'Id' in config:
            SubElement(config_node, 'Id', config['Id'])

        SubElement(config_node, NOTIFICATIONS_ARN_FIELDNAME_MAP[element_name],
                   config['Arn'])

        for event in config['Events']:
            SubElement(config_node, 'Event', event)

        filter_rules = config.get('Filter', {}).get(
            'Key', {}).get('FilterRules', [])
        if filter_rules:
            s3key_node = SubElement(SubElement(config_node, 'Filter'), 'S3Key')
            for filter_rule in filter_rules:
                filter_rule_node = SubElement(s3key_node, 'FilterRule')
                SubElement(filter_rule_node, 'Name', filter_rule['Name'])
                SubElement(filter_rule_node, 'Value', filter_rule['Value'])
    return node


def xml_marshal_delete_objects(keys):
    """
    Marshal Multi-Object Delete request body from object names.

    :param object_names: List of object keys to be deleted.
    :return: Serialized XML string for multi-object delete request body.
    """
    root = Element('Delete')

    # use quiet mode in the request - this causes the S3 Server to
    # limit its response to only object keys that had errors during
    # the delete operation.
    SubElement(root, 'Quiet', "true")

    # add each object to the request.
    for key in keys:
        version_id = None
        if not isinstance(key, basestring):
            version_id = key[1]
            key = key[0]

        element = SubElement(root, "Object")
        SubElement(element, "Key", key)
        if version_id:
            SubElement(element, "VersionId", version_id)

    return _get_xml_data(root)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from nose.tools import eq_

import mock
from minio import Minio
from minio.api import _DEFAULT_USER_AGENT

from .minio_mocks import MockConnection, MockResponse


class ListIncompleteUploadsTest(TestCase):
    @mock.patch('urllib3.PoolManager')
    def test_empty_list_uploads_test(self, mock_connection):
        mock_data = '''<?xml version="1.0"?>
<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
  <Bucket>golang</Bucket>
  <KeyMarker/>
  <UploadIdMarker/>
  <NextKeyMarker/>
  <NextUploadIdMarker/>
  <EncodingType/>
  <MaxUploads>1000</MaxUploads>
  <IsTruncated>false</IsTruncated>
  <Prefix/>
  <Delimiter/>
</ListMultipartUploadsResult>'''
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse('GET',
                         'https://localhost:9000/bucket/?prefix=&uploads=',
                         {'User-Agent': _DEFAULT_USER_AGENT}, 200,
                         content=mock_data)
        )
        client = Minio('localhost:9000')
        upload_iter = client._list_incomplete_uploads(
            'bucket', '', True, False)
        uploads = []
        for upload in upload_iter:
            uploads.append(upload)
        eq_(0, len(uploads))

    @mock.patch('urllib3.PoolManager')
    def test_list_uploads_works(self, mock_connection):
        mock_data = '''<?xml version="1.0"?>
<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
  <Bucket>golang</Bucket>
  <KeyMarker/>
  <UploadIdMarker/>
  <NextKeyMarker>keymarker</NextKeyMarker>
  <NextUploadIdMarker>uploadidmarker</NextUploadIdMarker>
  <EncodingType/>
  <MaxUploads>1000</MaxUploads>
  <IsTruncated>false</IsTruncated>
  <Upload>
    <Key>go1.4.2</Key>
    <UploadId>uploadid</UploadId>
    <Initiator>
      <ID/>
      <DisplayName/>
    </Initiator>
    <Owner>
      <ID/>
      <DisplayName/>
    </Owner>
    <StorageClass/>
    <Initiated>2015-05-30T14:43:35.349Z</Initiated>
  </Upload>
  <Upload>
    <Key>go1.5.0</Key>
    <UploadId>uploadid2</UploadId>
    <Initiator>
      <ID/>
      <DisplayName/>
    </Initiator>
    <Owner>
      <ID/>
      <DisplayName/>
    </Owner>
    <StorageClass/>
    <Initiated>2015-05-30T15:00:07.759Z</Initiated>
  </Upload>
  <Prefix/>
  <Delimiter/>
</ListMultipartUploadsResult>'''
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse('GET',
                         'https://localhost:9000/bucket/?delimiter=%2F&'
                         'prefix=&uploads=',
                         {'User-Agent': _DEFAULT_USER_AGENT},
                         200, content=mock_data)
        )

        client = Minio('localhost:9000')
        upload_iter = client._list_incomplete_uploads(
            'bucket', '', False, False)
        uploads = []
        for upload in upload_iter:
            uploads.append(upload)
        eq_(2, len(uploads))

    @mock.patch('urllib3.PoolManager')
    def test_list_multipart_uploads_works(self, mock_connection):
        mock_data1 = '''<?xml version="1.0"?>
<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
  <Bucket>golang</Bucket>
  <KeyMarker/>
  <UploadIdMarker/>
  <NextKeyMarker>keymarker</NextKeyMarker>
  <NextUploadIdMarker>uploadidmarker</NextUploadIdMarker>
  <EncodingType/>
  <MaxUploads>1000</MaxUploads>
  <IsTruncated>true</IsTruncated>
  <Upload>
    <Key>go1.4.2</Key>
    <UploadId>uploadid</UploadId>
    <Initiator>
      <ID/>
      <DisplayName/>
    </Initiator>
    <Owner>
      <ID/>
      <DisplayName/>
    </Owner>
    <StorageClass/>
    <Initiated>2015-05-30T14:43:35.349Z</Initiated>
  </Upload>
  <Upload>
    <Key>go1.5.0</Key>
    <UploadId>uploadid2</UploadId>
    <Initiator>
      <ID/>
      <DisplayName/>
    </Initiator>
    <Owner>
      <ID/>
      <DisplayName/>
    </Owner>
    <StorageClass/>
    <Initiated>2015-05-30T15:00:07.759Z</Initiated>
  </Upload>
  <Prefix/>
  <Delimiter/>
</ListMultipartUploadsResult>'''
        mock_data2 = '''<?xml version="1.0"?>
<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
  <Bucket>golang</Bucket>
  <KeyMarker/>
  <UploadIdMarker/>
  <NextKeyMarker/>
  <NextUploadIdMarker/>
  <EncodingType/>
  <MaxUploads>1000</MaxUploads>
  <IsTruncated>false</IsTruncated>
  <Upload>
    <Key>go1.4.2</Key>
    <UploadId>uploadid</UploadId>
    <Initiator>
      <ID/>
      <DisplayName/>
    </Initiator>
    <Owner>
      <ID/>
      <DisplayName/>
    </Owner>
    <StorageClass/>
    <Initiated>2015-05-30T14:43:35.349Z</Initiated>
  </Upload>
  <Upload>
    <Key>go1.5.0</Key>
    <UploadId>uploadid2</UploadId>
    <Initiator>
      <ID/>
      <DisplayName/>
    </Initiator>
    <Owner>
      <ID/>
      <DisplayName/>
    </Owner>
    <StorageClass/>
    <Initiated>2015-05-30T15:00:07.759Z</Initiated>
  </Upload>
  <Prefix/>
  <Delimiter/>
</ListMultipartUploadsResult>'''
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse('GET',
                         'https://localhost:9000/bucket/?prefix=&uploads=',
                         {'User-Agent': _DEFAULT_USER_AGENT}, 200,
                         content=mock_data1)
        )

        client = Minio('localhost:9000')
        upload_iter = client._list_incomplete_uploads(
            'bucket', '', True, False)
        uploads = []
        for upload in upload_iter:
            mock_server.mock_add_request(
                MockResponse('GET',
                             'https://localhost:9000/bucket/?'
                             'key-marker=keymarker&'
                             'prefix=&'
                             'upload-id-marker=uploadidmarker&uploads=',
                             {'User-Agent': _DEFAULT_USER_AGENT}, 200,
                             content=mock_data2)
            )
            uploads.append(upload)

        eq_(4, len(uploads))



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2017 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.fold_case_dict

This module implements a case insensitive dictionary.

:copyright: (c) 2017 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from .compat import PYTHON2


def _to_dict(value):
    """Create value to dictionary."""
    if not isinstance(value, dict):
        return value

    data = {}
    for key, val in value.items():
        data[key.lower()] = FoldCaseDict(val) if isinstance(val, dict) else val
    return data


class FoldCaseDict(dict):
    """Dictionary deals with case insensitive key."""

    def __init__(self, dictionary=None):
        super(FoldCaseDict, self).__init__()
        self._data = _to_dict(dictionary) if dictionary else {}

    def __getitem__(self, key):
        return self._data.__getitem__(key.lower())

    def __contains__(self, key):
        return self._data.__contains__(key.lower())

    def __setitem__(self, key, value):
        self._data.__setitem__(key.lower(), _to_dict(value))

    def __delitem__(self, key):
        self._data.__delitem__(key.lower())

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()

    def __eq__(self, other):
        if isinstance(other, dict):
            other = FoldCaseDict(other)
        elif isinstance(other, FoldCaseDict):
            pass
        else:
            raise NotImplementedError

        return self._data.__eq__(other._data)

    def __repr__(self):
        return self._data.__repr__()

    def get(self, key, default=None):
        return self._data.get(key.lower(), default)

    def has_key(self, key):
        """Test for the presence of key in the dictionary."""
        return key.lower() in self._data

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def iteritems(self):
        """Return an iterator over the dictionary’s (key, value) pairs."""
        if PYTHON2:
            return self._data.iteritems()  # pylint: disable=no-member
        return self._data.items()

    def iterkeys(self):
        """Return an iterator over the dictionary’s keys."""
        for k in self._data.keys():
            yield k

    def itervalues(self):
        """Return an iterator over the dictionary’s values."""
        for value in self._data.values():
            yield value

    def update(self, dictionary):
        if isinstance(dictionary, dict):
            dictionary = FoldCaseDict(dictionary)
        elif isinstance(dictionary, FoldCaseDict):
            pass
        else:
            raise TypeError

        self._data.update(dictionary._data)  # pylint: disable=protected-access

    def copy(self):
        return FoldCaseDict(self._data.copy())

    def clear(self):
        self._data = {}

    def pop(self, key):
        return self._data.pop(key.lower())

    def popitem(self):
        return self._data.popitem()



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015-2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-lines

"""
minio.error
~~~~~~~~~~~~~~~~~~~

This module provides custom exception classes for MinIO library
and API specific errors.

:copyright: (c) 2015, 2016, 2017 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""
from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError


class MinioError(Exception):
    """
    Base class for all exceptions

    :param message: User defined message.
    """

    def __init__(self, message, **kwargs):
        super(MinioError, self).__init__(**kwargs)
        self.message = message

    def __str__(self):
        return "{name}: message: {message}".format(
            name=self.__class__.__name__,
            message=self.message
        )


class InvalidEndpointError(MinioError):
    """
    InvalidEndpointError is raised when input endpoint URL is invalid.
    """


class InvalidBucketError(MinioError):
    """
    InvalidBucketError is raised when input bucket name is invalid.

    NOTE: Bucket names are validated based on Amazon S3 requirements.
    """


class InvalidArgumentError(MinioError):
    """
    InvalidArgumentError is raised when an unexpected
    argument is received by the callee.
    """


class InvalidSizeError(MinioError):
    """
    InvalidSizeError is raised when an unexpected size mismatch occurs.
    """


class InvalidXMLError(MinioError):
    """
    InvalidXMLError is raised when an unexpected XML tag or
    a missing tag is found during parsing.
    """


class MultiDeleteError(MinioError):
    """
    Represents an error raised when trying to delete an object in a
    Multi-Object Delete API call :class:`MultiDeleteError <MultiDeleteError>`

    :object_name: Object name that had a delete error.
    :error_code: Error code.
    :error_message: Error message.
    """

    def __init__(self, object_name, err_code, err_message):
        super(MultiDeleteError, self).__init__(err_message)
        self.object_name = object_name
        self.error_code = err_code
        self.error_message = err_message

    def __str__(self):
        string_format = ('<MultiDeleteError: object_name: {} error_code: {}'
                         ' error_message: {}>')
        return string_format.format(self.object_name,
                                    self.error_code,
                                    self.error_message)


class KnownResponseError(MinioError):
    """
    Common error responses listed in
    http://docs.aws.amazon.com/AmazonS3/latest/API/ErrorResponses.html#RESTErrorResponses
    """

    def __init__(self, response_error, **kwargs):
        super(KnownResponseError, self).__init__(
            message=self.message, **kwargs)
        self.response_error = response_error


class AccessDenied(KnownResponseError):
    """AccessDenied S3 error."""
    message = 'Access Denied'

    def __init__(self):
        super(AccessDenied, self).__init__(self.message)


class AccountProblem(KnownResponseError):
    """AcccountProblem S3 error."""
    message = ('There is a problem with your account that prevents the '
               'operation from completing successfully.')

    def __init__(self):
        super(AccountProblem, self).__init__(self.message)


class AmbiguousGrantByEmailAddress(KnownResponseError):
    """AmbiguousGrantByEmailAddress S3 error."""
    message = ('The email address you provided is associated with '
               'more than one account.')

    def __init__(self):
        super(AmbiguousGrantByEmailAddress, self).__init__(self.message)


class BadDigest(KnownResponseError):
    """BadDigest S3 error."""
    message = 'The Content-MD5 you specified did not match what we received.'

    def __init__(self):
        super(BadDigest, self).__init__(self.message)


class BucketAlreadyExists(KnownResponseError):
    """BucketAlreadyExists S3 error."""
    message = ('The requested bucket name is not available. The '
               'bucket namespace is shared by all users of the system. '
               'Please select a different name and try again.')

    def __init__(self):
        super(BucketAlreadyExists, self).__init__(self.message)


class BucketAlreadyOwnedByYou(KnownResponseError):
    """BucketAlreadyOwnedByYou S3 error."""
    message = ('Your previous request to create the named bucket '
               'succeeded and you already own it.')

    def __init__(self):
        super(BucketAlreadyOwnedByYou, self).__init__(self.message)


class BucketNotEmpty(KnownResponseError):
    """BucketNotEmpty S3 error."""
    message = 'The bucket you tried to delete is not empty.'

    def __init__(self):
        super(BucketNotEmpty, self).__init__(self.message)


class CredentialNotSupported(KnownResponseError):
    """CredentialNotSupported S3 error."""
    message = 'This request does not support credentials.'

    def __init__(self):
        super(CredentialNotSupported, self).__init__(self.message)


class CrossLocationLoggingProhibited(KnownResponseError):
    """CrossLocationLoggingProhibited S3 error."""
    message = ('Cross-location logging not allowed. Buckets in one '
               'geographic location cannot log information to a bucket '
               'in another location.')

    def __init__(self):
        super(CrossLocationLoggingProhibited, self).__init__(self.message)


class EntityTooSmall(KnownResponseError):
    """EntityTooSmall S3 error."""
    message = ('Your proposed upload is smaller than the minimum '
               'allowed object size.')

    def __init__(self):
        super(EntityTooSmall, self).__init__(self.message)


class EntityTooLarge(KnownResponseError):
    """EntityTooLarge S3 error."""
    message = 'Your proposed upload exceeds the maximum allowed object size.'

    def __init__(self):
        super(EntityTooLarge, self).__init__(self.message)


class ExpiredToken(KnownResponseError):
    """ExpiredToken S3 error."""
    message = 'The provided token has expired.'

    def __init__(self):
        super(ExpiredToken, self).__init__(self.message)


class IllegalVersioningConfigurationException(KnownResponseError):
    """IllegalVersioningConfigurationException S3 error."""
    message = ('Indicates that the versioning configuration specified '
               'in the request is invalid.')

    def __init__(self):
        super(IllegalVersioningConfigurationException, self).__init__(
            self.message)


class IncompleteBody(KnownResponseError):
    """IncompleteBody S3 error."""
    message = ('You did not provide the number of bytes specified by the '
               'Content-Length HTTP header')

    def __init__(self):
        super(IncompleteBody, self).__init__(self.message)


class IncorrectNumberOfFilesInPostRequest(KnownResponseError):
    """IncorrectNumberOfFilesInPostRequest S3 error."""
    message = 'POST requires exactly one file upload per request.'

    def __init__(self):
        super(IncorrectNumberOfFilesInPostRequest, self).__init__(self.message)


class InlineDataTooLarge(KnownResponseError):
    """InlineDataTooLarge S3 error."""
    message = 'Inline data exceeds the maximum allowed size.'

    def __init__(self):
        super(InlineDataTooLarge, self).__init__(self.message)


class InternalError(KnownResponseError):
    """InternalError S3 error."""
    message = 'We encountered an internal error. Please try again.'

    def __init__(self):
        super(InternalError, self).__init__(self.message)


class InvalidAccessKeyId(KnownResponseError):
    """InvalidAccessKeyId S3 error."""
    message = 'The access key Id you provided does not exist in our records.'

    def __init__(self):
        super(InvalidAccessKeyId, self).__init__(self.message)


class InvalidAddressingHeader(KnownResponseError):
    """InvalidAddressingHeader S3 error."""
    message = 'You must specify the Anonymous role.'

    def __init__(self):
        super(InvalidAddressingHeader, self).__init__(self.message)


class InvalidArgument(KnownResponseError):
    """InvalidArgument S3 error."""
    message = 'Invalid Argument'

    def __init__(self):
        super(InvalidArgument, self).__init__(self.message)


class InvalidBucketName(KnownResponseError):
    """InvalidBucketName S3 error."""
    message = 'The specified bucket is not valid.'

    def __init__(self):
        super(InvalidBucketName, self).__init__(self.message)


class InvalidBucketState(KnownResponseError):
    """InvalidBucketState S3 error."""
    message = 'The request is not valid with the current state of the bucket.'

    def __init__(self):
        super(InvalidBucketState, self).__init__(self.message)


class InvalidDigest(KnownResponseError):
    """InvalidDigest S3 error."""
    message = 'The Content-MD5 you specified is not valid.'

    def __init__(self):
        super(InvalidDigest, self).__init__(self.message)


class InvalidEncryptionAlgorithmError(KnownResponseError):
    """InvalidEncryptionAlgorithmError S3 error."""
    message = ('The encryption request you specified is not valid. '
               'The valid value is AES256.')

    def __init__(self):
        super(InvalidEncryptionAlgorithmError, self).__init__(self.message)


class InvalidLocationConstraint(KnownResponseError):
    """InvalidLocationConstraint S3 error."""
    message = 'The specified location constraint is not valid.'

    def __init__(self):
        super(InvalidLocationConstraint, self).__init__(self.message)


class InvalidObjectState(KnownResponseError):
    """InvalidObjectState S3 error."""
    message = 'The operation is not valid for the current state of the object.'

    def __init__(self):
        super(InvalidObjectState, self).__init__(self.message)


class InvalidPart(KnownResponseError):
    """InvalidPart S3 error."""
    message = ('One or more of the specified parts could not be found. '
               'The part might not have been uploaded, or the specified '
               'entity tag might not have matched the part\'s entity tag')

    def __init__(self):
        super(InvalidPart, self).__init__(self.message)


class InvalidPartOrder(KnownResponseError):
    """InvalidPartOrder S3 error."""
    message = ('The list of parts was not in ascending order.Parts list '
               'must specified in order by part number.')

    def __init__(self):
        super(InvalidPartOrder, self).__init__(self.message)


class InvalidPayer(KnownResponseError):
    """InvalidPayer S3 error."""
    message = 'All access to this object has been disabled.'

    def __init__(self):
        super(InvalidPayer, self).__init__(self.message)


class InvalidPolicyDocument(KnownResponseError):
    """InvalidPolicyDocument S3 error."""
    message = ('The content of the form does not meet the conditions '
               'specified in the policy document.')

    def __init__(self):
        super(InvalidPolicyDocument, self).__init__(self.message)


class InvalidRange(KnownResponseError):
    """InvalidRange S3 error."""
    message = 'The requested range cannot be satisfied.'

    def __init__(self):
        super(InvalidRange, self).__init__(self.message)


class InvalidRequest(KnownResponseError):
    """InvalidRequest S3 error."""
    message = 'Invalid Request'

    def __init__(self):
        super(InvalidRequest, self).__init__(self.message)


class InvalidSecurity(KnownResponseError):
    """InvalidSecurity S3 error."""
    message = 'The provided security credentials are not valid.'

    def __init__(self):
        super(InvalidSecurity, self).__init__(self.message)


class InvalidSOAPRequest(KnownResponseError):
    """InvalidSOAPRequest S3 error."""
    message = 'The SOAP request body is invalid.'

    def __init__(self):
        super(InvalidSOAPRequest, self).__init__(self.message)


class InvalidStorageClass(KnownResponseError):
    """InvalidStorageClass S3 error."""
    message = 'The storage class you specified is not valid.'

    def __init__(self):
        super(InvalidStorageClass, self).__init__(self.message)


class InvalidTargetBucketForLogging(KnownResponseError):
    """InvalidTargetBucketForLogging S3 error."""
    message = ('The target bucket for logging does not exist, '
               'is not owned by you, or does not have the appropriate '
               'grants for the log-delivery group.')

    def __init__(self):
        super(InvalidTargetBucketForLogging, self).__init__(self.message)


class InvalidToken(KnownResponseError):
    """InvalidToken S3 error."""
    message = 'The provided token is malformed or otherwise invalid.'

    def __init__(self):
        super(InvalidToken, self).__init__(self.message)


class InvalidURI(KnownResponseError):
    """InvalidURI S3 error."""
    message = 'Couldn\'t parse the specified URI.'

    def __init__(self):
        super(InvalidURI, self).__init__(self.message)


class KeyTooLong(KnownResponseError):
    """KeyTooLong S3 error."""
    message = 'Your key is too long.'

    def __init__(self):
        super(KeyTooLong, self).__init__(self.message)


class MalformedACLError(KnownResponseError):
    """MalformedACLError S3 error."""
    message = ('The XML you provided was not well-formed '
               'or did not validate against our published schema.')

    def __init__(self):
        super(MalformedACLError, self).__init__(self.message)


class MalformedPOSTRequest(KnownResponseError):
    """MalformedPOSTRequest S3 error."""
    message = ('The body of your POST request is not '
               'well-formed multipart/form-data.')

    def __init__(self):
        super(MalformedPOSTRequest, self).__init__(self.message)


class MalformedXML(KnownResponseError):
    """MalformedXML S3 error."""
    message = ("This happens when the user sends malformed xml (xml that "
               "doesn't conform to the published xsd) for the configuration.")

    def __init__(self):
        super(MalformedXML, self).__init__(self.message)


class MaxMessageLengthExceeded(KnownResponseError):
    """MaxMessageLengthExceeded S3 error."""
    message = 'Your request was too big.'

    def __init__(self):
        super(MaxMessageLengthExceeded, self).__init__(self.message)


class MaxPostPreDataLengthExceededError(KnownResponseError):
    """MaxPostPreDataLengthExceededError S3 error."""
    message = ('Your POST request fields preceding the '
               'upload file were too large.')

    def __init__(self):
        super(MaxPostPreDataLengthExceededError, self).__init__(self.message)


class MetadataTooLarge(KnownResponseError):
    """MetadataTooLarge S3 error."""
    message = 'Your metadata headers exceed the maximum allowed metadata size.'

    def __init__(self):
        super(MetadataTooLarge, self).__init__(self.message)


class MethodNotAllowed(KnownResponseError):
    """MethodNotAllowed S3 error."""
    message = 'The specified method is not allowed against this resource'

    def __init__(self):
        super(MethodNotAllowed, self).__init__(self.message)


class MissingAttachment(KnownResponseError):
    """MissingAttachment S3 error."""
    message = 'A SOAP attachment was expected, but none were found.'

    def __init__(self):
        super(MissingAttachment, self).__init__(self.message)


class MissingContentLength(KnownResponseError):
    """MissingContentLength S3 error."""
    message = 'You must provide the Content-Length HTTP header.'

    def __init__(self):
        super(MissingContentLength, self).__init__(self.message)


class MissingRequestBodyError(KnownResponseError):
    """MissingRequestBodyError S3 error."""
    message = ('This happens when the user sends an empty xml document '
               'as a request. The error message is, "Request body is empty."')

    def __init__(self):
        super(MissingRequestBodyError, self).__init__(self.message)


class MissingSecurityElement(KnownResponseError):
    """MissingSecurityElement S3 error."""
    message = 'The SOAP 1.1 request is missing a security element.'

    def __init__(self):
        super(MissingSecurityElement, self).__init__(self.message)


class MissingSecurityHeader(KnownResponseError):
    """MissingSecurityHeader S3 error."""
    message = 'Your request is missing a required header.'

    def __init__(self):
        super(MissingSecurityHeader, self).__init__(self.message)


class NoLoggingStatusForKey(KnownResponseError):
    """NoLoggingStatusForKey S3 error."""
    message = ('There is no such thing as a logging '
               'status subresource for a key.')

    def __init__(self):
        super(NoLoggingStatusForKey, self).__init__(self.message)


class NoSuchBucket(KnownResponseError):
    """NoSuchBucket S3 error."""
    message = 'The specified bucket does not exist.'

    def __init__(self):
        super(NoSuchBucket, self).__init__(self.message)


class NoSuchKey(KnownResponseError):
    """NoSuchKey S3 error."""
    message = 'The specified key does not exist.'

    def __init__(self):
        super(NoSuchKey, self).__init__(self.message)


class NoSuchLifecycleConfiguration(KnownResponseError):
    """NoSuchLifecycleConfiguration S3 error."""
    message = 'The lifecycle configuration does not exist.'

    def __init__(self):
        super(NoSuchLifecycleConfiguration, self).__init__(self.message)


class NoSuchUpload(KnownResponseError):
    """NoSuchUpload S3 error."""
    message = ('The specified multipart upload does not exist. '
               'The upload ID might be invalid, or the multipart '
               'upload might have been aborted or completed.')

    def __init__(self):
        super(NoSuchUpload, self).__init__(self.message)


class NoSuchVersion(KnownResponseError):
    """NoSuchVersion S3 error."""
    message = ('Indicates that the version ID specified in the '
               'request does not match an existing version.')

    def __init__(self):
        super(NoSuchVersion, self).__init__(self.message)


class APINotImplemented(KnownResponseError):
    """NotImplemented S3 error."""
    message = ('A header you provided implies functionality '
               'that is not implemented.')

    def __init__(self):
        super(APINotImplemented, self).__init__(self.message)


class NotSignedUp(KnownResponseError):
    """NotSignedUp S3 error."""
    message = 'Your account is not signed up.'

    def __init__(self):
        super(NotSignedUp, self).__init__(self.message)


class NoSuchBucketPolicy(KnownResponseError):
    """NoSuchBucketPolicy S3 error."""
    message = 'The specified bucket does not have a bucket policy.'

    def __init__(self):
        super(NoSuchBucketPolicy, self).__init__(self.message)


class OperationAborted(KnownResponseError):
    """OperationAborted S3 error."""
    message = ('A conflicting conditional operation is currently in '
               'progress against this resource. Try again.')

    def __init__(self):
        super(OperationAborted, self).__init__(self.message)


class PermanentRedirect(KnownResponseError):
    """PermanentRedirect S3 error."""
    message = ('The bucket you are attempting to access must be addressed '
               'using the specified endpoint. Send all future requests '
               'to this endpoint.')

    def __init__(self):
        super(PermanentRedirect, self).__init__(self.message)


class PreconditionFailed(KnownResponseError):
    """PreconditionFailed S3 error."""
    message = 'At least one of the preconditions you specified did not hold.'

    def __init__(self):
        super(PreconditionFailed, self).__init__(self.message)


class Redirect(KnownResponseError):
    """Redirect S3 error."""
    message = 'Temporary redirect.'

    def __init__(self):
        super(Redirect, self).__init__(self.message)


class RestoreAlreadyInProgress(KnownResponseError):
    """RestoreAlreadyInProgress S3 error."""
    message = 'Object restore is already in progress.'

    def __init__(self):
        super(RestoreAlreadyInProgress, self).__init__(self.message)


class RequestIsNotMultiPartContent(KnownResponseError):
    """RequestIsNotMultiPartContent S3 error."""
    message = 'Bucket POST must be of the enclosure-type multipart/form-data.'

    def __init__(self):
        super(RequestIsNotMultiPartContent, self).__init__(self.message)


class RequestTimeout(KnownResponseError):
    """RequestTimeout S3 error."""
    message = ('Your socket connection to the server was not read '
               'from or written to within the timeout period.')

    def __init__(self):
        super(RequestTimeout, self).__init__(self.message)


class RequestTimeTooSkewed(KnownResponseError):
    """RequestTimeTooSkewed S3 error."""
    message = ("The difference between the request time and the "
               "server's time is too large.")

    def __init__(self):
        super(RequestTimeTooSkewed, self).__init__(self.message)


class RequestTorrentOfBucketError(KnownResponseError):
    """RequestTorrentOfBucketError S3 error."""
    message = 'Requesting the torrent file of a bucket is not permitted.'

    def __init__(self):
        super(RequestTorrentOfBucketError, self).__init__(self.message)


class SignatureDoesNotMatch(KnownResponseError):
    """SignatureDoesNotMatch S3 error."""
    message = ('The request signature we calculated does not match the '
               'signature you provided.')

    def __init__(self):
        super(SignatureDoesNotMatch, self).__init__(self.message)


class ServiceUnavailable(KnownResponseError):
    """ServiceUnavailable S3 error."""
    message = "Service unavailable. Retry again."

    def __init__(self):
        super(ServiceUnavailable, self).__init__(self.message)


class SlowDown(KnownResponseError):
    """SlowDown S3 error."""
    message = 'Reduce your request rate.'

    def __init__(self):
        super(SlowDown, self).__init__(self.message)


class TemporaryRedirect(KnownResponseError):
    """TemporaryRedirect S3 error."""
    message = 'You are being redirected to the bucket while DNS updates.'

    def __init__(self):
        super(TemporaryRedirect, self).__init__(self.message)


class TokenRefreshRequired(KnownResponseError):
    """TokenRefreshRequired S3 error."""
    message = 'The provided token must be refreshed.'

    def __init__(self):
        super(TokenRefreshRequired, self).__init__(self.message)


class TooManyBuckets(KnownResponseError):
    """TooManyBuckets S3 error."""
    message = 'You have attempted to create more buckets than allowed.'

    def __init__(self):
        super(TooManyBuckets, self).__init__(self.message)


class UnexpectedContent(KnownResponseError):
    """UnexpectedContent S3 error."""
    message = 'This request does not support content.'

    def __init__(self):
        super(UnexpectedContent, self).__init__(self.message)


class UnresolvableGrantByEmailAddress(KnownResponseError):
    """UnresolvableGrantByEmailAddress S3 error."""
    message = ('The email address you provided does not match any account '
               'on record.')

    def __init__(self):
        super(UnresolvableGrantByEmailAddress, self).__init__(self.message)


class UserKeyMustBeSpecified(KnownResponseError):
    """UserKeyMustBeSpecified S3 error."""
    message = ('The bucket POST must contain the specified field name. '
               'If it is specified, check the order of the fields.')

    def __init__(self):
        super(UserKeyMustBeSpecified, self).__init__(self.message)


_KNOWN_ERRORS = {
    'AccessDenied': AccessDenied,
    'AccountProblem': AccountProblem,
    'AmbiguousGrantByEmailAddress': AmbiguousGrantByEmailAddress,
    'BadDigest': BadDigest,
    'BucketAlreadyExists': BucketAlreadyExists,
    'BucketAlreadyOwnedByYou': BucketAlreadyOwnedByYou,
    'BucketNotEmpty': BucketNotEmpty,
    'CredentialNotSupported': CredentialNotSupported,
    'CrossLocationLoggingProhibited': CrossLocationLoggingProhibited,
    'EntityTooSmall': EntityTooSmall,
    'EntityTooLarge': EntityTooLarge,
    'ExpiredToken': ExpiredToken,
    'IllegalVersioningConfigurationException':
    IllegalVersioningConfigurationException,
    'IncompleteBody': IncompleteBody,
    'IncorrectNumberOfFilesInPostRequest': IncorrectNumberOfFilesInPostRequest,
    'InlineDataTooLarge': InlineDataTooLarge,
    'InternalError': InternalError,
    'InvalidAccessKeyId': InvalidAccessKeyId,
    'InvalidAddressingHeader': InvalidAddressingHeader,
    'InvalidArgument': InvalidArgument,
    'InvalidBucketName': InvalidBucketName,
    'InvalidBucketState': InvalidBucketState,
    'InvalidDigest': InvalidDigest,
    'InvalidEncryptionAlgorithmError': InvalidEncryptionAlgorithmError,
    'InvalidLocationConstraint': InvalidLocationConstraint,
    'InvalidObjectState': InvalidObjectState,
    'InvalidPart': InvalidPart,
    'InvalidPartOrder': InvalidPartOrder,
    'InvalidPayer': InvalidPayer,
    'InvalidPolicyDocument': InvalidPolicyDocument,
    'InvalidRange': InvalidRange,
    'InvalidRequest': InvalidRequest,
    'InvalidSecurity': InvalidSecurity,
    'InvalidSOAPRequest': InvalidSOAPRequest,
    'InvalidStorageClass': InvalidStorageClass,
    'InvalidTargetBucketForLogging': InvalidTargetBucketForLogging,
    'InvalidToken': InvalidToken,
    'InvalidURI': InvalidURI,
    'KeyTooLong': KeyTooLong,
    'MalformedACLError': MalformedACLError,
    'MalformedPOSTRequest': MalformedPOSTRequest,
    'MalformedXML': MalformedXML,
    'MaxMessageLengthExceeded': MaxMessageLengthExceeded,
    'MaxPostPreDataLengthExceededError': MaxPostPreDataLengthExceededError,
    'MetadataTooLarge': MetadataTooLarge,
    'MethodNotAllowed': MethodNotAllowed,
    'MissingAttachment': MissingAttachment,
    'MissingContentLength': MissingContentLength,
    'MissingRequestBodyError': MissingRequestBodyError,
    'MissingSecurityElement': MissingSecurityElement,
    'MissingSecurityHeader': MissingSecurityHeader,
    'NoLoggingStatusForKey': NoLoggingStatusForKey,
    'NoSuchBucket': NoSuchBucket,
    'NoSuchKey': NoSuchKey,
    'NoSuchLifecycleConfiguration': NoSuchLifecycleConfiguration,
    'NoSuchUpload': NoSuchUpload,
    'NoSuchVersion': NoSuchVersion,
    'NotImplemented': APINotImplemented,
    'NotSignedUp': NotSignedUp,
    'NoSuchBucketPolicy': NoSuchBucketPolicy,
    'OperationAborted': OperationAborted,
    'PermanentRedirect': PermanentRedirect,
    'PreconditionFailed': PreconditionFailed,
    'Redirect': Redirect,
    'RestoreAlreadyInProgress': RestoreAlreadyInProgress,
    'RequestIsNotMultiPartContent': RequestIsNotMultiPartContent,
    'RequestTimeout': RequestTimeout,
    'RequestTimeTooSkewed': RequestTimeTooSkewed,
    'RequestTorrentOfBucketError': RequestTorrentOfBucketError,
    'SignatureDoesNotMatch': SignatureDoesNotMatch,
    'ServiceUnavailable': ServiceUnavailable,
    'SlowDown': SlowDown,
    'TemporaryRedirect': TemporaryRedirect,
    'TokenRefreshRequired': TokenRefreshRequired,
    'TooManyBuckets': TooManyBuckets,
    'UnexpectedContent': UnexpectedContent,
    'UnresolvableGrantByEmailAddress': UnresolvableGrantByEmailAddress,
    'UserKeyMustBeSpecified': UserKeyMustBeSpecified,
}


class ResponseError(MinioError):
    """
    ResponseError is raised when an API call doesn't succeed.
    raises :exc:`ResponseError` accordingly.

    :param response: Response from http client :class:`urllib3.HTTPResponse`.
    """

    def __init__(self, response, method, bucket_name=None,
                 object_name=None):
        super(ResponseError, self).__init__(message='')
        # initialize parameter fields
        self._response = response
        self._xml = response.data
        self.method = method
        self.bucket_name = bucket_name
        self.object_name = object_name
        # initialize all ResponseError fields
        self.code = ''
        # Amz headers
        self.request_id = ''
        self.host_id = ''
        self.region = ''

        # handle the error
        self._handle_error_response(bucket_name)

    def get_exception(self):
        """
        Gets the error exception derived from the initialization of
        an ErrorResponse object

        :return: The derived exception or ResponseError exception
        """
        return _KNOWN_ERRORS.get(self.code, self)

    def _handle_error_response(self, bucket_name=None):
        """
        Sets error response uses xml body if available, otherwise
        relies on HTTP headers.
        """
        if self._response.data:
            self._set_from_response_body()
        else:
            self._set_from_response_reason(bucket_name)

    def _set_from_response_body(self):
        """
        Sets all the error response fields with a valid response body.
           Raises :exc:`ValueError` if invoked on a zero length body.

        :param bucket_name: Optional bucket name resource at which error
           occurred.
        :param object_name: Option object name resource at which error
           occurred.
        """
        if not self._response.data:
            raise ValueError('response data has no body.')
        try:
            root = ElementTree.fromstring(self._response.data)
        except (ParseError, AttributeError, ValueError, TypeError) as error:
            raise InvalidXMLError('"Error" XML is not parsable. '
                                  'Message: {0}'.format(error))

        # Deal with namespaced response from sts
        tag_prefix = "{https://sts.amazonaws.com/doc/2011-06-15/}"
        if tag_prefix not in root.tag:
            tag_prefix = ""

        attr_dict = {
            tag_prefix + 'Code': 'code',
            tag_prefix + 'BucketName': 'bucket_name',
            tag_prefix + 'Key': 'object_name',
            tag_prefix + 'Message': 'message',
            tag_prefix + 'RequestId': 'request_id',
            tag_prefix + 'HostId': 'host_id'
        }
        for attribute in root.iter():
            attr = attr_dict.get(attribute.tag)
            if attr:
                setattr(self, attr, attribute.text)
        # Set amz headers.
        self._set_amz_headers()

    def _set_from_response_reason(self, bucket_name):
        """
        Sets all the error response fields from response headers.
        """
        status_dict = {
            301: lambda: ('PermanentRedirect', self._response.reason),
            307: lambda: ('Redirect', self._response.reason),
            400: lambda: ('BadRequest', self._response.reason),
            403: lambda: ('AccessDenied', self._response.reason),
            404: lambda: (
                (self.code, self.message) if not bucket_name else
                ('NoSuchKey', self._response.reason) if self.object_name else
                ('NoSuchBucket', self._response.reason)
            ),
            405: lambda: ('MethodNotAllowed', self._response.reason),
            409: lambda: ('Conflict',
                          'The bucket you tried to delete is not empty.'),
            500: lambda: ('InternalError', 'Internal Server Error.'),
            501: lambda: ('MethodNotAllowed', self._response.reason),
        }

        func = status_dict.get(self._response.status, lambda: (
            'UnknownException', self._response.reason))
        self.code, self.message = func()

        # Set amz headers.
        self._set_amz_headers()

    def _set_amz_headers(self):
        """
        Sets x-amz-* error response fields from response headers.
        """
        if self._response.headers:
            self.host_id = self._response.headers.get('x-amz-id-2',
                                                      self.host_id)
            self.request_id = self._response.headers.get('x-amz-request-id',
                                                         self.request_id)
            # 'x-amz-bucket-region' is a new undocumented field.
            self.region = self._response.headers.get('x-amz-bucket-region',
                                                     self.region)

    def __str__(self):
        return ('ResponseError: code: {0}, message: {1},'
                ' bucket_name: {2}, object_name: {3}, request_id: {4},'
                ' host_id: {5}, region: {6}').format(self.code,
                                                     self.message,
                                                     self.bucket_name,
                                                     self.object_name,
                                                     self.request_id,
                                                     self.host_id,
                                                     self.region)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016, 2017, 2018, 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from nose.tools import eq_

from minio.definitions import UploadPart
from minio.select.options import (CSVInput, CSVOutput, InputSerialization,
                                  OutputSerialization, RequestProgress,
                                  SelectObjectOptions)
from minio.xml_marshal import (marshal_complete_multipart,
                               xml_marshal_bucket_constraint,
                               xml_marshal_select)


class GenerateRequestTest(TestCase):
    def test_generate_bucket_constraint(self):
        expected_string = (b'<CreateBucketConfiguration '
                           b'xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
                           b'<LocationConstraint>region</LocationConstraint>'
                           b'</CreateBucketConfiguration>')
        actual_string = xml_marshal_bucket_constraint('region')
        eq_(expected_string, actual_string)

    def test_generate_complete_multipart_upload(self):
        expected_string = (b'<CompleteMultipartUpload '
                           b'xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
                           b'<Part><PartNumber>1</PartNumber>'
                           b'<ETag>"a54357aff0632cce46d942af68356b38"</ETag>'
                           b'</Part>'
                           b'<Part><PartNumber>2</PartNumber>'
                           b'<ETag>"0c78aef83f66abc1fa1e8477f296d394"</ETag>'
                           b'</Part>'
                           b'<Part><PartNumber>3</PartNumber>'
                           b'<ETag>"acbd18db4cc2f85cedef654fccc4a4d8"</ETag>'
                           b'</Part>'
                           b'</CompleteMultipartUpload>')

        etags = [
            UploadPart('bucket', 'object', 'upload_id', 1,
                       'a54357aff0632cce46d942af68356b38',
                       None, 0),
            UploadPart('bucket', 'object', 'upload_id', 2,
                       '0c78aef83f66abc1fa1e8477f296d394',
                       None, 0),
            UploadPart('bucket', 'object', 'upload_id', 3,
                       'acbd18db4cc2f85cedef654fccc4a4d8',
                       None, 0),
        ]
        actual_string = marshal_complete_multipart(etags)
        eq_(expected_string, actual_string)

    def test_xml_marshal_select(self):
        expected_string = (b'<SelectObjectContentRequest>'
                           b'<Expression>select * from s3object</Expression>'
                           b'<ExpressionType>SQL</ExpressionType>'
                           b'<InputSerialization>'
                           b'<CompressionType>NONE</CompressionType>'
                           b'<CSV><FileHeaderInfo>USE</FileHeaderInfo>'
                           b'<RecordDelimiter>\n</RecordDelimiter>'
                           b'<FieldDelimiter>,</FieldDelimiter>'
                           b'<QuoteCharacter>"</QuoteCharacter>'
                           b'<QuoteEscapeCharacter>"</QuoteEscapeCharacter>'
                           b'<Comments>#</Comments>'
                           b'<AllowQuotedRecordDelimiter>false'
                           b'</AllowQuotedRecordDelimiter></CSV>'
                           b'</InputSerialization>'
                           b'<OutputSerialization><CSV>'
                           b'<QuoteFields>ASNEEDED</QuoteFields>'
                           b'<RecordDelimiter>\n</RecordDelimiter>'
                           b'<FieldDelimiter>,</FieldDelimiter>'
                           b'<QuoteCharacter>"</QuoteCharacter>'
                           b'<QuoteEscapeCharacter>"</QuoteEscapeCharacter>'
                           b'</CSV></OutputSerialization>'
                           b'<RequestProgress>'
                           b'<Enabled>true</Enabled>'
                           b'</RequestProgress>'
                           b'</SelectObjectContentRequest>')

        options = SelectObjectOptions(
            expression="select * from s3object",
            input_serialization=InputSerialization(
                compression_type="NONE",
                csv=CSVInput(file_header_info="USE",
                             record_delimiter="\n",
                             field_delimiter=",",
                             quote_character='"',
                             quote_escape_character='"',
                             comments="#",
                             allow_quoted_record_delimiter=False),
            ),
            output_serialization=OutputSerialization(
                csv=CSVOutput(quote_fields="ASNEEDED",
                              record_delimiter="\n",
                              field_delimiter=",",
                              quote_character='"',
                              quote_escape_character='"')
            ),
            request_progress=RequestProgress(
                enabled=True
            )
        )
        actual_string = xml_marshal_select(options)
        eq_(expected_string, actual_string)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015-2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
minio.signer
~~~~~~~~~~~~~~~

This module implements all helpers for AWS Signature version '4' support.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import collections
import hashlib
import hmac
from datetime import datetime

from .compat import queryencode, urlsplit
from .error import InvalidArgumentError
from .fold_case_dict import FoldCaseDict
from .helpers import get_sha256_hexdigest

# Signature version '4' algorithm.
_SIGN_V4_ALGORITHM = 'AWS4-HMAC-SHA256'

# Hardcoded S3 header value for X-Amz-Content-Sha256
_UNSIGNED_PAYLOAD = u'UNSIGNED-PAYLOAD'

# Hardcoded service name for pre-signed S3 urls
_PRESIGNED_SERVICE_NAME = "s3"

# Default service name for all signature
_DEFAULT_SERVICE_NAME = "s3"


def post_presign_signature(date, region, secret_key, policy_str):
    """
    Calculates signature version '4' for POST policy string.

    :param date: datetime formatted date.
    :param region: region of the bucket for the policy.
    :param secret_key: Amazon S3 secret access key.
    :param policy_str: policy string.
    :return: hexlified sha256 signature digest.
    """
    signing_key = generate_signing_key(date, region, secret_key)
    signature = hmac.new(signing_key, policy_str.encode('utf-8'),
                         hashlib.sha256).hexdigest()

    return signature


def presign_v4(method, url, credentials,
               region=None, headers=None, expires=None,
               response_headers=None, request_date=None):
    """
    Calculates signature version '4' for regular presigned URLs.

    :param method: Method to be presigned examples 'PUT', 'GET'.
    :param url: URL to be presigned.
    :param credentials: Credentials object with your AWS s3 account info.
    :param region: region of the bucket, it is optional.
    :param headers: any additional HTTP request headers to
       be presigned, it is optional.
    :param expires: final expiration of the generated URL. Maximum is 7days.
    :param response_headers: Specify additional query string parameters.
    :param request_date: the date of the request.
    """

    # Validate input arguments.
    if not credentials.get().access_key or not credentials.get().secret_key:
        raise InvalidArgumentError('Invalid access_key and secret_key.')

    region = region or 'us-east-1'
    headers = headers or {}
    expires = expires or '604800'
    request_date = request_date or datetime.utcnow()

    # If a sha256sum is known, add to headers to include with signature
    content_hash_hex = _UNSIGNED_PAYLOAD
    for k in headers:
        if k.lower() == 'x-amz-content-sha256':
            content_hash_hex = headers[k] or _UNSIGNED_PAYLOAD
            del headers[k]
            break

    parsed_url = urlsplit(url)
    host = remove_default_port(parsed_url)
    headers['Host'] = host
    iso8601_date = request_date.strftime("%Y%m%dT%H%M%SZ")

    headers_to_sign = headers
    # Construct queries.
    query = {}
    query['X-Amz-Algorithm'] = _SIGN_V4_ALGORITHM
    query['X-Amz-Credential'] = generate_credential_string(
        credentials.get().access_key,
        request_date,
        region,
        _PRESIGNED_SERVICE_NAME,
    )
    query['X-Amz-Date'] = iso8601_date
    query['X-Amz-Expires'] = str(expires)
    if credentials.get().session_token:
        query['X-Amz-Security-Token'] = credentials.get().session_token

    signed_headers = get_signed_headers(headers_to_sign)
    query['X-Amz-SignedHeaders'] = ';'.join(signed_headers)

    if response_headers:
        query.update(response_headers)

    # URL components.
    url_components = [parsed_url.geturl()]
    ordered_query = collections.OrderedDict(sorted(query.items()))
    query_components = []
    for component_key in ordered_query:
        single_component = [component_key, '=']
        if ordered_query[component_key]:
            single_component.append(queryencode(ordered_query[component_key]))
        query_components.append(''.join(single_component))

    query_string = '&'.join(query_components)
    if query_string:
        url_components.append('?')
        url_components.append(query_string)
    new_url = ''.join(url_components)
    # new url constructor block ends.
    new_parsed_url = urlsplit(new_url)

    canonical_request = generate_canonical_request(method,
                                                   new_parsed_url,
                                                   headers_to_sign,
                                                   signed_headers,
                                                   content_hash_hex)
    string_to_sign = generate_string_to_sign(request_date, region,
                                             canonical_request,
                                             _PRESIGNED_SERVICE_NAME)
    signing_key = generate_signing_key(request_date, region,
                                       credentials.get().secret_key)
    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                         hashlib.sha256).hexdigest()
    new_parsed_url = urlsplit(new_url + "&X-Amz-Signature="+signature)
    return new_parsed_url.geturl()


def get_signed_headers(headers):
    """
    Get signed headers.

    :param headers: input dictionary to be sorted.
    """
    return sorted([h.lower().strip() for h in headers])


def sign_v4(method, url, region, headers=None,
            credentials=None,
            content_sha256=None,
            request_datetime=None,
            service_name=_DEFAULT_SERVICE_NAME
            ):
    """
    Signature version 4.

    :param method: HTTP method used for signature.
    :param url: Final url which needs to be signed.
    :param region: Region should be set to bucket region.
    :param headers: Optional headers for the method.
    :param credentials: Optional Credentials object with your AWS s3 account
                        info.
    :param content_sha256: Optional body sha256.
    :param request_datetime: Optional request date/time
    :param service_name: Optional service to sign request for (defaults to S3)
    """

    # If no access key or secret key is provided return headers.
    if not credentials.get().access_key or not credentials.get().secret_key:
        return headers

    headers = headers or FoldCaseDict()
    region = region or 'us-east-1'

    parsed_url = urlsplit(url)
    secure = parsed_url.scheme == 'https'
    if secure and not content_sha256:
        content_sha256 = _UNSIGNED_PAYLOAD
    content_sha256 = content_sha256 or get_sha256_hexdigest('')

    host = remove_default_port(parsed_url)
    headers['Host'] = host

    request_datetime = request_datetime or datetime.utcnow()

    headers['X-Amz-Date'] = request_datetime.strftime("%Y%m%dT%H%M%SZ")
    headers['X-Amz-Content-Sha256'] = content_sha256
    if credentials.get().session_token:
        headers['X-Amz-Security-Token'] = credentials.get().session_token

    headers_to_sign = headers

    signed_headers = get_signed_headers(headers_to_sign)
    canonical_req = generate_canonical_request(method,
                                               parsed_url,
                                               headers_to_sign,
                                               signed_headers,
                                               content_sha256
                                               )

    string_to_sign = generate_string_to_sign(request_datetime, region,
                                             canonical_req, service_name)
    signing_key = generate_signing_key(
        request_datetime, region, credentials.get().secret_key, service_name)
    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                         hashlib.sha256).hexdigest()

    authorization_header = generate_authorization_header(
        credentials.get().access_key, request_datetime, region, signed_headers,
        signature, service_name)

    headers['Authorization'] = authorization_header
    return headers


def generate_canonical_request(method, parsed_url, headers, signed_headers,
                               content_sha256):
    """
    Generate canonical request.

    :param method: HTTP method.
    :param parsed_url: Parsed url is input from :func:`urlsplit`
    :param headers: HTTP header dictionary.
    :param content_sha256: Content sha256 hexdigest string.
    """
    # Should not encode ~. Decode it back if present.
    parsed_url_path = parsed_url.path.replace("%7E", "~")
    parsed_url_query = parsed_url.query.replace("%7E", "~")
    lines = [method, parsed_url_path, parsed_url_query]

    # Headers added to canonical request.
    header_lines = []
    for header in signed_headers:
        value = headers[header.title()]
        value = str(value).strip()
        header_lines.append(header + ':' + value)

    lines = lines + header_lines
    lines.append('')

    lines.append(';'.join(signed_headers))
    lines.append(content_sha256)
    return '\n'.join(lines)


def generate_string_to_sign(date, region, canonical_request,
                            service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate string to sign.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param canonical_request: Canonical request generated previously.
    :param service_name: Service to scope request for.
    """
    formatted_date_time = date.strftime("%Y%m%dT%H%M%SZ")

    canonical_request_hasher = hashlib.sha256()
    canonical_request_hasher.update(canonical_request.encode('utf-8'))
    canonical_request_sha256 = canonical_request_hasher.hexdigest()
    scope = generate_scope_string(date, region, service_name)

    return '\n'.join([_SIGN_V4_ALGORITHM,
                      formatted_date_time,
                      scope,
                      canonical_request_sha256])


def generate_signing_key(date, region, secret_key,
                         service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate signing key.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param secret_key: Secret access key.
    :param service_name: The signing key is scoped to a service e.g. s3
    """
    formatted_date = date.strftime("%Y%m%d")

    key1_string = 'AWS4' + secret_key
    key1 = key1_string.encode('utf-8')
    key2 = hmac.new(key1, formatted_date.encode('utf-8'),
                    hashlib.sha256).digest()
    key3 = hmac.new(key2, region.encode('utf-8'), hashlib.sha256).digest()
    key4 = hmac.new(key3, service_name.encode(
        'utf-8'), hashlib.sha256).digest()

    return hmac.new(key4, 'aws4_request'.encode('utf-8'),
                    hashlib.sha256).digest()


def generate_scope_string(date, region, service_name):
    """
    Generate scope string.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param service_name: Service for scope string, e.g., "s3".
    """
    formatted_date = date.strftime("%Y%m%d")
    scope = '/'.join([formatted_date,
                      region,
                      service_name,
                      'aws4_request'])
    return scope


def generate_credential_string(access_key, date, region,
                               service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate credential string.

    :param access_key: Server access key.
    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param service_name: Service to scope credentials to.
    """
    return access_key + '/' + generate_scope_string(date, region, service_name)


def generate_authorization_header(access_key, date, region,
                                  signed_headers, signature,
                                  service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate authorization header.

    :param access_key: Server access key.
    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param signed_headers: Signed headers.
    :param signature: Calculated signature.
    :param service_name: Optional service to sign request for.
    """
    signed_headers_string = ';'.join(signed_headers)
    credential = generate_credential_string(
        access_key, date, region, service_name)
    auth_header = [_SIGN_V4_ALGORITHM, 'Credential=' + credential + ',',
                   'SignedHeaders=' + signed_headers_string + ',',
                   'Signature=' + signature]
    return ' '.join(auth_header)


def remove_default_port(parsed_url):
    """Remove default port in URL."""
    default_ports = {
        'http': 80,
        'https': 443
    }
    if any(parsed_url.scheme == scheme and parsed_url.port == port
           for scheme, port in default_ports.items()):
        # omit default port (i.e. 80 or 443)
        host = parsed_url.hostname
    else:
        host = parsed_url.netloc
    return host



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# AssumeRoleProvider will call the Simple Token Service (STS) to retrieve
# temporary credentials.
#
# - You can't call AssumeRole as a root user on either MinIO or AWS.
#   For MinIO add a non-root user using the minio client `mc`:
#
#     mc admin user add myminio YOUR-ACCESSKEYID YOUR-SECRETACCESSKEY
#   On AWS you will need an IAM user with the sts:AssumeRole action allowed,
#   and a target role.
# - The credentials will be valid for between 15 minutes and 12 hours.
# - An access policy can be applied to the temporary credentials. The
#   resulting permissions are the intersection of the role's existing policy
#   and the optionally provided policy. You cannot grant more permissions than
#   those allowed by the policy of the role that is being assumed.
# - YOUR-ACCESSKEYID and YOUR-SECRETACCESSKEY are
#   dummy values, please replace them with original values.
# - To use MinIO with AWS, pass the function - `get_assume_role_creds`
#   as a lambda to the AssumeRoleProvider

from minio import Minio
from minio.credentials import AssumeRoleProvider, Credentials

user_client = Minio(
    's3.amazonaws.com',
    access_key='YOUR-ACCESSKEYID',
    secret_key='YOUR-SECRETKEY',
)

_RESTRICTED_UPLOAD_POLICY = """{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Action": [
        "s3:*"
      ],
      "Effect": "Allow",
      "Resource": [
        "arn:aws:s3:::my-bucket/*"
      ],
      "Sid": "Upload-access-to-specific-bucket-only"
    }
  ]
}
"""

provider = AssumeRoleProvider(
    lambda: user_client.get_assume_role_creds(_RESTRICTED_UPLOAD_POLICY),
)

client = Minio(
    's3.amazonaws.com',
    credentials=Credentials(provider),
)
